import os
from datetime import datetime
import numpy as np
import json
import matplotlib.pyplot as plt
import torch
import rsp.common.console as console
import pickle as pkl

class Run():
    def __init__(self, id = None, moving_average_epochs = 1000):
        if id is None:
            self.id = datetime.now().strftime('%Y%m%d%H%M%S%f')
            self.data = {}
        else:
            self.id = id
            self.__load__()
            
        self.moving_average_epochs = moving_average_epochs
        self.__init_run_dir__()
    
    def append(self, key:str, phase:str, value):
        if not key in self.data:
            self.data[key] = {}
        if not phase in self.data[key]:
            self.data[key][phase] = {
                'val': [],
                'avg': []
            }
        if np.isnan(value):
            if len(self.data[key][phase]['val']) > 0:
                value = self.data[key][phase]['val'][-1]
            else:
                value = 0.

        self.data[key][phase]['val'].append(value)
        self.data[key][phase]['avg'].append(np.average(self.data[key][phase]['val'][-self.moving_average_epochs:]))

    def plot(self):
        if not os.path.isdir(f'runs/{self.id}/plot'):
            os.mkdir(f'runs/{self.id}/plot')

        for key in self.data:
            key_str = key.replace('_', ' ')
            fname = f'runs/{self.id}/plot/{key}.jpg'

            cmap = plt.get_cmap('tab20b')
            colors = cmap(np.linspace(0, 1, len(self.data[key])))

            for i, phase in enumerate(self.data[key]):
                if len(self.data[key][phase]['val']) == 0:
                    continue
                plt.plot(self.data[key][phase]['val'], color=colors[i], alpha=0.3)
                plt.plot(self.data[key][phase]['avg'], label=phase, color=colors[i])

            plt.title(key_str)
            plt.xlabel('episode')
            plt.ylabel(key_str)
            plt.minorticks_on()
            plt.grid(which='minor', color='lightgray', linewidth=0.2)
            plt.grid(which='major', linewidth=.6)
            plt.legend()
            plt.savefig(fname)
            plt.close()

    def save(self):
        self.__init_run_dir__()
        with open(f'runs/{self.id}/data.json', 'w') as f:
            json.dump(self.data, f)

    def get_val(self, key:str, phase:str):
        return self.data[key][phase]['val'][-1]
    
    def get_avg(self, key:str, phase:str):
        return self.data[key][phase]['avg'][-1]
    
    def len(self):
        l = 0
        for key in self.data:
            for phase in self.data[key]:
                l_temp = len(self.data[key][phase]['val'])
                if l_temp > l:
                    l = l_temp
        return l

    def __init_run_dir__(self):
        if not os.path.isdir('runs'):
            os.mkdir('runs')
        self.directory = f'runs/{self.id}'
        if not os.path.isdir(self.directory):
            os.mkdir(self.directory)
        self.plot_directory = f'{self.directory}/plot'
        if not os.path.isdir(self.plot_directory):
            os.mkdir(self.plot_directory)

    def __load__(self):
        if not os.path.isfile(f'runs/{self.id}/data.json'):
            self.data = {}
        else:
            with open(f'runs/{self.id}/data.json', 'r') as f:
                self.data = json.load(f)

    def save_state_dict(self, state_dict, fname = 'state_dict.pt'):
        self.__init_run_dir__()
        with open(f'runs/{self.id}/{fname}', 'wb') as f:
            torch.save(state_dict, f)

    def load_state_dict(self, model:torch.nn.Module, fname = 'state_dict.pt'):
        if not os.path.isfile(f'runs/{self.id}/{fname}'):
            console.warn(f'File runs/{self.id}/{fname} not found.')
            return
        with open(f'runs/{self.id}/{fname}', 'rb') as f:
            model.load_state_dict(torch.load(f))

    def save_model(self, model:torch.nn.Module, fname = 'model.pkl'):
        self.__init_run_dir__()
        with open(f'runs/{self.id}/{fname}', 'wb') as f:
            pkl.dump(model, f)

    def load_model(self, fname = 'model.pkl'):
        if not os.path.isfile(f'runs/{self.id}/{fname}'):
            console.warn(f'File runs/{self.id}/{fname} not found.')
            return
        with open(f'runs/{self.id}/{fname}', 'rb') as f:
            model = pkl.load(f)
        return model


from pathlib import Path
from googledriver import download_folder
import zipfile
import importlib
import sys
import rsp.common.console as console
import torch

TUC_ActionPrediction_model004 = 'TUC/ActionPrediction/Model4'

URL = 'https://drive.google.com/drive/folders/1ulNnPqg-5wvenRl2CuJMxMMcaiYfHjQ9?usp=share_link'

def __download_model_folder__():
    download_folder(URL)
    pass

def load_model(model_id:str, force_reload:bool = False) -> torch.nn.Module:
    zip_file = Path(f'Model/{model_id}.zip')
    if not zip_file.exists() or force_reload:
        __download_model_folder__()
        console.success(f'Downloaded models from {URL}.')

    assert zip_file.exists(), f'File {zip_file} does not exist.'

    with zipfile.ZipFile(zip_file, 'r') as zip_ref:
        zip_ref.extractall(zip_file.parent)
    pass

    class_name = model_id.split('/')[-1]

    model_def_py = zip_file.parent.joinpath(class_name).joinpath('model.py')
    spec = importlib.util.spec_from_file_location("model", model_def_py)
    foo = importlib.util.module_from_spec(spec)
    sys.modules["module.name"] = foo
    spec.loader.exec_module(foo)
    model:torch.nn.Module = foo.__dict__[class_name]()

    model_state_dict_file = zip_file.parent.joinpath(class_name).joinpath('state_dict.pt')
    with open(model_state_dict_file, 'rb') as f:
        model.load_state_dict(torch.load(f))

    return model

if __name__ == '__main__':
    model = load_model(TUC_ActionPrediction_model004)


import torch
import torch.nn.functional as F
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sn
import pandas as pd
import cv2 as cv

def confusion_matrix(Y, T):
    num_classes = T.shape[1]
    classes_Y = torch.argmax(Y, dim=1)
    classes_T = torch.argmax(T, dim=1)

    cm = torch.zeros((num_classes, num_classes), dtype=torch.int64)

    for cy, ct in zip(classes_Y, classes_T):
        cm[ct, cy] += 1

    return cm

def confusion(Y, T):
    """ Returns the confusion matrix for the values in the `prediction` and `truth`
        tensors, i.e. the amount of positions where the values of `prediction`
        and `truth` are
        - 1 and 1 (True Positive)
        - 1 and 0 (False Positive)
        - 0 and 0 (True Negative)
        - 0 and 1 (False Negative)
        """
    confusion_vector = Y / T
    # Element-wise division of the 2 tensors returns a new tensor which holds a
    # unique value for each case:
    #   1     where prediction and truth are 1 (True Positive)
    #   inf   where prediction is 1 and truth is 0 (False Positive)
    #   nan   where prediction and truth are 0 (True Negative)
    #   0     where prediction is 0 and truth is 1 (False Negative)
    true_positives = torch.sum(confusion_vector == 1).item()
    false_positives = torch.sum(confusion_vector == float('inf')).item()
    true_negatives = torch.sum(torch.isnan(confusion_vector)).item()
    false_negatives = torch.sum(confusion_vector == 0).item()
    return true_positives, false_positives, true_negatives, false_negatives

def TP(Y, T):
    assert torch.all(Y <= 1) and torch.all(Y >= 0), 'Expected 0 <= Y <= 1'
    
    Y_pos = Y >= 0.5
    T_pos = T >= 0.5

    mask = torch.bitwise_and(Y_pos, T_pos)
    tp = mask.sum().item()
    return tp

def TN(Y, T):
    assert torch.all(Y <= 1) and torch.all(Y >= 0), 'Expected 0 <= Y <= 1'
    
    Y_neg = Y < 0.5
    T_neg = T < 0.5

    mask = torch.bitwise_and(Y_neg, T_neg)
    tn = mask.sum().item()
    return tn

def FP(Y, T):
    assert torch.all(Y <= 1) and torch.all(Y >= 0), 'Expected 0 <= Y <= 1'
    
    Y_pos = Y >= 0.5
    T_neg = T < 0.5

    mask = torch.bitwise_and(Y_pos, T_neg)
    fp = mask.sum().item()
    return fp

def FN(Y, T):
    assert torch.all(Y <= 1) and torch.all(Y >= 0), 'Expected 0 <= Y <= 1'
    
    Y_neg = Y < 0.5
    T_pos = T >= 0.5

    mask = torch.bitwise_and(Y_neg, T_pos)
    fn = mask.sum().item()
    return fn

def FPR(Y, T):
    fp = FP(Y, T)
    tn = TN(Y, T)
    if fp + tn == 0:
        return np.nan
    fpr = fp / (fp + tn)
    return fpr

def TPR(Y, T):
    tp = TP(Y, T)
    fn = FN(Y, T)
    if tp + fn == 0:
        return np.nan
    tpr = tp / (tp + fn)
    return tpr

def F1_Score(Y, T):
    tp = TP(Y, T)
    fp = FP(Y, T)
    fn = FN(Y, T)
    if tp + fp + fn == 0:
        return np.nan
    f1score = 2 * tp / (2 * tp + fp + fn)
    return f1score

def top_k_accuracy(Y:torch.Tensor, T:torch.Tensor, k:int):
    top_k_prediction = torch.argsort(Y, dim=1, descending=True)[:, :k]
    top_1_target = torch.argmax(T, dim=1)

    tp = 0
    for y, t in zip(top_k_prediction, top_1_target):
        if t in y:
            tp += 1
        pass

    acc = tp / T[:, 0].numel()
    return acc

def top_1_accuracy(Y:torch.Tensor, T:torch.Tensor):
    return top_k_accuracy(Y, T, 1)

def top_3_accuracy(Y:torch.Tensor, T:torch.Tensor):
    return top_k_accuracy(Y, T, 3)

def plot_confusion_matrix(
        confusion_matrix:torch.Tensor,
        labels = None,
        cmap:str = 'Blues',
        xlabel:str = 'Predicted label',
        ylabel:str = 'True label',
        title:str = 'Confusion Matrix',
        plt_show:bool = False,
        save_file_name:str = None) -> np.array:
    if labels is None:
        labels = [f'Class {i+1}' for i in range(confusion_matrix.shape[0])]

    df_cm = pd.DataFrame(confusion_matrix, index = [i for i in labels],
                  columns = [i for i in labels])
    fig = plt.figure(figsize = (10,7))

    sn.heatmap(df_cm, annot=True, cmap=cmap)
    plt.xticks(rotation=45, ha="right")
    plt.yticks(rotation=0, ha="right")
    if title is not None:
        plt.title(title)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)

    if plt_show:
        plt.show()

    fig.canvas.draw()

    s, (width, height) = fig.canvas.print_to_buffer()

    # Option 2a: Convert to a NumPy array.
    img = np.fromstring(s, np.uint8).reshape((height, width, 4))
    img = cv.cvtColor(img, cv.COLOR_BGR2RGB)

    if save_file_name is not None:
        cv.imwrite(save_file_name, img)

    plt.close()

    return img

if __name__ == '__main__':
    Y = torch.tensor([
        [0.1, 0.1, 0.8],
        [0.95, 0.03, 0.02],
        [0.05, 0.9, 0.05]
    ])
    T = torch.tensor([
        [0, 0, 1],
        [1, 0, 0],
        [0, 1, 0]
    ])

    tp = TP(Y, T)
    tn = TN(Y, T)
    fp = FP(Y, T)
    fn = FN(Y, T)

    f1 = F1_Score(Y, T)

    epsilon = 0.2
    num_elements = 1000
    num_classes = 10

    T = []
    for i in range(num_elements):
        true_class = torch.randint(0, num_classes, (1,))
        t = F.one_hot(true_class, num_classes=num_classes)
        T.append(t)
    T = torch.cat(T)

    dist = torch.normal(T.float(), 1.5)
    Y = torch.argmax(dist, dim=1)
    Y = F.one_hot(Y, num_classes=num_classes)

    conf_m = confusion_matrix(Y, T)
    print(conf_m)

    labels = []
    for a in range(num_classes):
        a_str = str(a)
        while len(a_str) < 3:
            a_str = '0' + a_str
        a_str = 'A' + a_str
        labels.append(a_str)

    img = plot_confusion_matrix(conf_m, labels=labels, plt_show=True)
    img = plot_confusion_matrix(conf_m, labels=labels, plt_show=True)